/*

+----------------------------------------------------------------------------+
!                                                                            !
!   yDec 1.0  14.Nov.2001                                                    !
!                                                                            !
!   This sourcecode is released to the public domain - no rights reserved    !
!   More info - also about the yEncoding - at:  www.yenc.org                 !
!                                                                            !
!   If you make a portation to another OS then please notify me (TIA)        !
!          archiver@i3w.com    (Questions welcome)                           !
!                                                                            !
!   Juergen Helbing - infstar@infostar.de                                    !
!                                                                            !
!                                                                            !
!   The 'offline' multipart feature has been implemented as a 'BackPack'.    !
!   I'm not really happy about it - it is simply Q&D (quick and dirty).      !
!                                                                            !
+----------------------------------------------------------------------------+


******************************************************************************
This Code if for the Linux/Unix (actually should work for anything that has
GCC and a command shell. MacOSX, BeOS, QNX, AtheOS, FreeBSD, etc. *SHOULD* 
work fine with this version. All features a pretty close to the old version
(most of the code is the same, just reworked to function in a Unix 
environment).

      - Michael Gaskins
        mbgaski@cs.clemson.edu

******************************************************************************

*/

#include "ydec.h"

/*------------------------------------------------------------------------------*/

void CrcAdd(int c) {
  unsigned long ch1,ch2,cc;  
  
  /* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
  /* for (i = 0; i < size; i++) */
  /*      crccode = crc32Tab[(int) ((crccode) ^ (buf[i])) & 0xff] ^  */
  /*                (((crccode) >> 8) & 0x00FFFFFFL); */
  /*   return(crccode); */
  
  
  cc= (c) & 0x000000ffL;
  ch1=(crc_val ^ cc) & 0xffL;
  ch1=crc_tab[ch1];
  ch2=(crc_val>>8L) & 0xffffffL;  // Correct version
  crc_val=ch1 ^ ch2;
  crc_anz++;
}

/*------------------------------------------------------------------------------*/

//  hex_to_ulong makes a conversion of an 8 character CRC to an unsigned long value
//  strtol fails for CRCs which start with 89ABCDEF  !
unsigned long hex_to_ulong(char * text) {
  unsigned long res;
  unsigned char c;
  
  if (text==NULL) 
    return(-1);

  res=0;
 loop:
  c=*text; 
  text++;
  if ((c>='0')&(c<='9')) {
    res=(res<<4)+((long)(c-48) & 0x0F);
    goto loop;
  }
  if ((c>='A')&(c<='F')) {
    res=(res<<4)+((long)(c-55) & 0x0F);
    goto loop;
  }
  if ((c>='a')&(c<='f')) {
    res=(res<<4)+((long)(c-87) & 0x0F);
    goto loop;
  }
  return res;
}

/*------------------------------------------------------------------------------*/

// ad_fgetscr  is a functions which simulates 'fgets' but removes CRLF at the line end
char * ad_fgetscr(char * buffer, int maxlen, FILE * fp) {
  char *cp, *dp;
  
  cp=fgets(buffer,maxlen,fp);
  if (cp==NULL) 
    return(NULL);
  
  dp=(char*)strrchr(buffer,'\n');   // Eliminate CRLF
  if (dp) 
    *dp=0;
  dp=(char*)strrchr(buffer,'\r');
  if (dp) 
    *dp=0;
  
  return cp;
  
}

/*------------------------------------------------------------------------------*/

void CrcInit() {
  crc_val = -1L ;
  crc_anz=0L;
}

/*------------------------------------------------------------------------------*/

int AddPart(char * srcname,long y_begin, long y_end) {
  FILE *fSrc, *fDes;
  char desname[260], line[1024];
  long a, e, flen;
  int copyrest, resparts;
  char *cp;
  
  if (aDsp) 
    print("AddPart (%s) [%ld - %ld]\r\n",srcname,y_begin,y_end);
  
  copyrest=0;
  resparts=0;
  
  sprintf(desname,"%s.des",srcname);
  
  fSrc=fopen(srcname,"rb");
  if (fSrc==NULL) {
    eprint("AddPart:  Part Info File not found (%s)\r\n",srcname);
    errors++;
    return -1;
  }
  flen = filelength(fileno(fSrc));
  if (flen==0) { // Already complete  
    return -9;
  }
  fDes = fopen(desname,"wb");
  if (fDes==NULL) {
    eprint("AddPart: Cannot create temporary part file (%s) [reason: %d]\r\n",desname,errno);
    fclose(fSrc);
    errors++;
    return -2;
  }
 loop:
  cp=fgets(line,1000,fSrc);
  if (cp == NULL) 
    goto eof1;

  cp = (char*) strrchr(line,'\n');

  if (cp) 
    *cp=0;

  cp = (char*) strrchr(line,'\r');

  if (cp) 
    *cp=0;
  
  // print("%s\r\n",line);
  
  if (copyrest) {  // Analysis done - copy rest & done.
    
    fprintf(fDes,"%s\r\n",line);
    resparts++;
    goto loop;
  }
  
  a=atol(line);
  if (a==0) {
    eprint("Corrupt part-info file (%s)\r\n",line);
    errors++;
    fclose(fSrc); fclose(fDes); return(-3);
  }
  cp = (char*) strchr(line,',');
  
  if (cp==NULL) {
    eprint("Corrupt part-info file (%s)\r\n",line);
    errors++;
    fclose(fSrc); fclose(fDes); return(-4);
  }
  e=atol(cp+1);
  //	print("a:%6d, e:%6d\r\n",a,e);
  
  // ----- Case Analysis -----
  //	.............[yb--------ye]      (The new, incoming block)
  //	1...a****e                       (The old, missing block)
  //	2...a********-----e              (***) Newly written missing block
  //	3...a********--------------****e
  //	.............[yb--------ye]
  //	4............a-----e
  // 5............a------------e
  //	6............a-------------****e
  //	.............[yb--------ye]
  //	7............................a***e
  //	.............[yb--------ye]
  //	8................a----e
  //	9................a---------****e  
  
  if (a<y_begin) {  // A previous section was missed
    
      if (e<y_begin) { // A previous section was missed entirely
	
	// Case-1
	fprintf(fDes,"%ld,%ld\r\n",a,e);  // --> Keep that entry
	resparts++;
	goto loop;
      }
      if (e<=y_end) {  // Rduce the old section
	// Case-2
	fprintf(fDes,"%ld,%ld\r\n",a,y_begin-1);  // --> Reduce that entry
	resparts++;
	goto loop;
      }
      // The new section is splitting the old missing section
      // Case-3
      fprintf(fDes,"%ld,%ld\r\n",a,y_begin-1);  // --> First part
      fprintf(fDes,"%ld,%ld\r\n",y_end+1,e);  // --> First part
      resparts++; resparts++;
      goto loop;
  }
  
  if (a == y_begin) {  // New part matches this segment
    if (e < y_end) {   // Just a smaller section was missed
      
      // case 4
      goto loop;  // Skip this and analyse next
    }
    if (e == y_end) {  // This missing section was exactly -> remove it & done.
      
      // case 5
      copyrest=1; 
      goto loop;
    }
    if (e > y_end) { // The first part was found
      
      // case-6
      fprintf(fDes,"%ld,%ld\r\n",y_end+1,e);  // Write the rest
      resparts++;
      copyrest=1; 
      goto loop;
    }
  }
  
  if (a > y_begin) {
    if (a > y_end) {  // the found section completely in front of this one.
      
      // Case-7
      fprintf(fDes,"%ld,%ld\r\n",a,e);  // Rewrite it & copy the rest
      resparts++;
      copyrest=1; goto loop;
    }
    if (e <= y_end) { // The section was completely found
      
      // Case-8
      goto loop;  // Analyse the rest
    }
    // The first part of the missing section was found
    // case-9
    fprintf(fDes,"%ld,%ld\r\n",y_end+1,e);  // Reduce it & copy the rest
    resparts++;
    copyrest=1; 
    goto loop;
  }
  
 eof1:
  fclose(fSrc); 
  fclose(fDes);
  unlink(srcname);
  
  rename(desname,srcname);
  
  return resparts;
}

/*------------------------------------------------------------------------------*/

long filelength (int fd) { 
  struct stat *st; 
  
  if (fstat (fd, st) == -1) { 
    exit (1); 
  } 
  return st->st_size; 
} 

/*------------------------------------------------------------------------------*/

int yDecode(FILE * fOut, FILE * fIn, long y_line, long y_size,int y_part) {
  unsigned char srcbuf[4100], desbuf[4100];
  unsigned char *srcp, *desp, c;
  int deslen, srclen, id;
  char * cp;
  long decolen;

  //	long sumlen;
  //	int partnr;
  
  unsigned long crc32;
  char name[260];
  int esize;
  unsigned char * partbuf;
  unsigned char * partptr;
  long partsize, partfree, wlen, flen;
  
  if (aDsp) 
    print("yDecoder started...\r\n");
	
  // sumlen=0; partnr=1;
  
 part_start:
  CrcInit();  // Analyse only CRC per part
  decolen=0;
  deslen=0; 
  desp=desbuf;
  
  if (y_part) {  // This is a multipart message !
    
    cp=ad_fgetscr((char*) srcbuf,4097,fIn);  // fgets especially with ad_length
    if (cp==NULL) {
      
      eprint("Unexpected eof in yEncoded file\r\n");
      errors++;
      return(1);
    }
    
    if (aDsp) 
      print("=ypart-line:  %s\r\n",srcbuf);
    
    if (strncmp((char*) srcbuf,"=ypart ",7)) {
      eprint("Missing =ypart line in yEncoded multipart message\r\n");
      errors++;
      return(2);
    }
    cp = (char*) strstr((char*)srcbuf,"end=");
    if (cp==NULL) {
      eprint("Missing end= in yEncoded multipart message\r\n");
      errors++;
      return(2);
    }
    y_end=atol(cp+4);
    
    cp = (char*) strstr((char*)srcbuf,"begin=");
    if (cp==NULL) {
      eprint("Missing begin= in yEncoded multipart message\r\n");
      errors++;
      return(2);
    }
    y_begin=atol(cp+6);
    
    if (aDsp) print("part-begin: %ld\r\n",y_begin);
    if (aDsp) print("part-end  : %ld\r\n",y_end);
    
    partbuf=malloc(y_end-y_begin+10);  // Allocate a buffer for the part
    partptr=partbuf;
    partsize=y_end-y_begin+1;
    partfree=partsize;
  }
  
 loop:
  
  cp=ad_fgetscr((char*) srcbuf,4097,fIn);  // fgets especially with ad_length
  if (cp==NULL) {
    eprint("Unexpected eof in yEncoded file\r\n");
    errors++;
    return(1);
  }
  if (strncmp((char*) srcbuf,"=yend ",6)==0) {
    if (aDsp) print("Endline (%d bytes): %s\r\n",decolen,srcbuf);
    goto end_of_file;
  }
  if (strncmp((char*) srcbuf,"=ybegin ",8)==0) {
    eprint("Unexpected =ybegin in yEncoded file\r\n");
    errors++;
    return(1);
  }
  srclen=strlen((char*)srcbuf);
  if (srclen<y_line) {
    if (aDsp) 
      print("Last line.\r\n");
  }
  srcp=srcbuf;
  
 loop2:
  c=*srcp; 
  srcp++;
  if (c==0) {
    goto loop;  // End of line reached
  }
  if (c == '=')  // The escape character comes in
    {
      c=*srcp; srcp++;
      if (c==0) return(2); // Last char cannot be escape char !
      c=(unsigned char)(c-64);
    }
  c=(unsigned char)(c-42);  // Subtract the secret number
  *desp=c; desp++; deslen++;	decolen++;
  
  CrcAdd(c);
  
  if (deslen>=4096) {
    
    if (y_part) {  // MultiPart --> to the partbuffer !
      
      if (deslen>partfree) {
	eprint("Partial message corrupt: longer than (end-begin)!\r\n");
	errors++;
	return(11);
      }
      memcpy(partptr,desbuf,deslen);
      partptr=partptr+deslen;
      partfree=partfree-deslen;
    }
    else {    // Single part --> directly to target file
      
      id=fwrite(desbuf,deslen,1,fOut);
      if (id != 1) {
	eprint("Error in writing decoded file (code=%d)\r\n",errno);
	errors++;
	return(3);
      }
    }
    deslen=0; desp=desbuf;
  }
  goto loop2;
  
 end_of_file:
  
  
  if (deslen>0) {  // Empty the last buffer
    if (y_part) {
      if (deslen>partfree) {
	eprint("Partial message corrupt: longer than (end-begin)!\r\n");
	errors++;
	return(11);
      }
      memcpy(partptr,desbuf,deslen);
      //partptr=partptr+deslen;
      //partfree=partfree-deslen;
    }
    else {    // Single part --> directly to target file
      
      id=fwrite(desbuf,deslen,1,fOut);
      if (id != 1) {
	eprint("Error in writing decoded file (code=%d)\r\n",errno);
	errors++;
	return(3);
      }
    }
    

    
    //		id=fwrite(desbuf,deslen,1,fOut);
    //		if (id != 1)
    //			{
    //			eprint("Error in writing decoded file (code=%d)\r\n",errno);
    //			return(4);
    //			}
  }
  
  cp = (char*) strstr((char*) srcbuf,"size=");  // Compare the decoded size to the =yend size
  if (cp) {
    esize=atoi(cp+5);
    if (esize != decolen) {
      sprintf(name,"%s(size=%ld)",attname,decolen);
      strcpy(attname,name);
      eprint("Corrupt yEnc binary - endsize mismatch (%s%s)\r\n",attname,attext);
      errors++; errfiles++;
      return(0);
    }
  }

  
  // Check the srcbuf for the CRC
  if (y_part==0) {
    cp = (char*) strstr((char*)srcbuf,"crc32=");
    if (cp) {
      crc32=hex_to_ulong((char*)(cp+6));
      ad_rescrc=crc_val ^ 0xFFFFFFFFl;
      if (aDsp) print("Included CRC: $%08lx - calculated CRC: $%08lx\r\n",crc32,ad_rescrc);
    }
  }
  else {
    cp = (char*) strstr((char*)srcbuf,"pcrc32=");
    if (cp) {
      crc32=hex_to_ulong((char*)(cp+7));
      ad_rescrc=crc_val ^ 0xFFFFFFFFl;
      if (aDsp) print("Included CRC: $%08lx - calculated CRC: $%08lx\r\n",crc32,ad_rescrc);
    }
  }
  
  if (cp!=NULL) {
    if (crc32 != ad_rescrc) {
      sprintf(name,"%s(crc=%08lx)",attname,ad_rescrc);
      strcpy(attname,name);
      eprint("Corrupt yEnc binary - CRC mismatch (%s%s)\r\n",attname,attext);
      errors++; errfiles++;
      return(0);
    }
  }
  
  if (y_part==0) {  // Single message
    
    if ((y_part==0) & (decolen != y_size)) {
      // eprint("Y-Decoder: Size mismatch - file corrupt.\r\n");
      
      sprintf(name,"%s(len=%ld)",attname,decolen);
      strcpy(attname,name);
      eprint("Corrupt yEnc binary - size mismatch (%s%s)\r\n",attname,attext);
      errors++; errfiles++;
      return(0);
    }
    
    ad_reslen=decolen;
    ad_rescrc=crc_val ^ 0xFFFFFFFFl;
    
    if (aDsp) 
      print("yDecoder: Job done. %ld bytes written. CRC: $%08lx \r\n",decolen,ad_rescrc);
    return(0);
  }
  
  // Multipart message
  
  if ((y_part>0) & (decolen != (y_end-y_begin+1))) {
    eprint("yDecoder: Part size mismatch - file corrupt.\r\n");
    errors++; 
    errparts++;
    return(6);
  }

  // ----- Special handling for external decoder
  
  // Now write the decoded part to the target file !
  
  id=fseek(fOut,y_begin-1,0);
  if (id) {
    eprint("Cannot write a part (fseek failed) [reason:%d]\r\n",errno);
    errors++;
    return(12);
  }
  flen=decolen;
  partptr=partbuf;
  
  if (aDsp) 
    print("Write part to target file (start: %ld, size: %ld)\r\n",y_begin-1,flen);
  
  while (flen>0) {
    wlen=flen;
    if (wlen>8192) 
      wlen=8192;
    id=fwrite(partptr,wlen,1,fOut);
    
    if (id != 1) {
      eprint("Cannot write a part (fwrite failed) [reason:%d]\r\n",errno);
      errors++;
      return(12);
    }
    partptr=partptr+wlen;
    flen=flen-wlen;
  }
  
  free(partbuf);  // Give the part-buffer back !
  
  return(0);

}

/*------------------------------------------------------------------------------*/

int DecodeFile(char * fname) {
  FILE * fIn;
  char line[8200];
  char * cp;
  char filename[260];
  FILE * fOut;
  int id;
  long flen,wlen;
  char filename2[260];
  
  fIn=fopen(fname,"rb");
  if (fIn==NULL) {
    eprint("Cannot open sourcefile (%s) [reason:%d]\r\n",fname,errno);
    errors++;
    return(1);
  }
 loop:
  cp=fgets(line,8192,fIn);    // Read source line
  if (cp==NULL) 
    goto eof1;
  
  cp = (char*) strrchr(line,'\n');     // Eliminate CRLF
  if (cp) 
    *cp=0;
  cp = (char*) strrchr(line,'\r');
  if (cp) 
    *cp=0;
  
  if (strncmp(line,"=ybegin ",8)) 
    goto loop;
  
  
  if (aDsp) 
    print("\r\nTrigger: %s\r\n",line);
  
  // Start of a section found
  
  cp = (char*) strstr(line,"name=");
  if (cp==NULL) {
    eprint("'name=' not found in =ybegin line. (%s)\r\n",line);
    errors++;
    goto loop;  // Error - filename not found
  }
  strcpy(attname,cp+5);  // Store the filename
  *cp=0; // throw away the filename
  
  cp = (char*) strstr(line,"size=");
  if (cp==NULL) {
    eprint("'size=' not found in =ybegin line. (%s)\r\n",line);
    errors++;
    goto loop;  // Error - size not found
  }
  y_size=atol(cp+5);
  
  cp = (char*) strstr(line,"line=");
  if (cp==NULL) {
    eprint("'line=' not found in =ybegin line. (%s)\r\n",line);
    errors++;
    goto loop;  // Error - linelength not found
  }
  y_line=atol(cp+5);
  
  y_part=0;
  cp = (char*) strstr(line,"part=");   // Check if this is a multipart message
  if (cp) {
    y_part=atol(cp+5);
    //			if (y_part != 1) goto loop;   // MUST start with the first part for proper decoding
  }
  
  // Now we can decode the file -- let's assume that this is a valid file
  
  attext[0]=0;
  cp = (char*) strrchr(attname,'.');
  if (cp) {
    strcpy(attext,cp); *cp=0;
  }
  
  // sprintf(filename,"%s\\decode.tmp",datadir);
  sprintf(filename,"%sdecode.tmp",ad_despath);
  
  // ----- Special Handling for offline decoder (missing parts or mixed sequence)
  if (y_part) {
    sprintf(filename ,"%s%s%s(%ld).tmp",ad_despath,attname,attext,y_size);
    sprintf(filename2,"%s%s%s(%ld).dec",ad_despath,attname,attext,y_size);
  }
  // ----------------------------------------------------------------------
  
  
  if (aDsp) 
    print("DesFileName = (%s%s)\r\n",attname,attext);
  
  // 27-03-2001: Set the filename just in case we cannot decode it
  sprintf(ad_desname,"%s\\%s%s",ad_despath,attname,attext);
  // sprintf(ad_desname,"%s%s",attname,attext);
  
  // filelen=0;
  
  
  // Check if the target file exists already
  if (y_part) {
    
    fOut=fopen(filename2,"rb");
    if (fOut) {
      flen=filelength(fileno(fOut));
      fclose(fOut);
      if (flen==0) {
	print("Part#%3d duplicate for finished %s \r\n",y_part,ad_desname);
	goto loop;
      }
    }
    
    fOut=fopen(filename,"r+b");
    if (fOut) {
      flen=filelength(fileno(fOut));
      if (flen==y_size) {
	goto already_open;
      }
      else {
	fclose(fOut);
	eprint("Size mismatch on target multipart file !(act:%ld, new:%ld) (%s)\r\n",flen,y_size,filename);
	errors++;
	goto abortus;
      }
    }
    else { // First time: Create this file & the missing index
      
      fOut=fopen(filename2,"wb");
      if (fOut==NULL) {
	eprint("Cannot create part-info-file (%s)\r\n",filename2);
	errors++;
	goto abortus;
      }
      fprintf(fOut,"%ld,%ld\r\n",1,y_size);  // Everything is missing !
      fclose(fOut);
    }
  }
  
  fOut=fopen(filename,"w+b");
  if (fOut==NULL) {  // Possibly already exists
    
    eprint("AutoDecode: Cannot create DesFile (%s)\r\n",filename);
    errors++;
    fclose(fIn);
    return(-1);
    // goto loop;
  }
  
  if (y_part) {  // Now create the entire file !

    flen=y_size;
    memset(line,255,8192);
    while (flen>0) {
      wlen=flen;
      if (wlen>8192) 
	wlen=8192;
      id=fwrite(line,wlen,1,fOut);
      if (id != 1) {
	eprint("Could not create the target file (disk full?)\r\n");
	fclose(fOut);
	fclose(fIn);
	errors++;
	return(-2);
      }
      flen=flen-wlen;
    }
  }
 already_open:
  
  // Now decode that file !
  
  id=yDecode(fOut,fIn,y_line,y_size,y_part);
  if (id) {
    eprint("yDecode failed (reason: %d)\r\n",id);
    goto abortus;
  }
  sprintf(ad_desname,"%s%s%s",ad_despath,attname,attext); // Create new target name
  if (aDsp) print("yDecode successful\r\n");
  fclose(fOut);
  
  // ----- Special handling for offline multipart
  
  if (y_part) {  // Now update the 'missing part file'
    
    id=AddPart(filename2,y_begin,y_end);
    if (id==0) { // All parts done
      
      print("Complete multipart --> %s\r\n",ad_desname);
      goto rename_it;
    }
    if (id>0) {
      print("Part#%3d: %s --> %s\r\n",y_part,fname,ad_desname);
    }
    if (id==(-9)) {
      print("Part#%3d of finished %s \r\n",y_part,ad_desname);
    }
  }
  if (y_part) 
    goto loop;
  // --------------------------------------------
  
 rename_it:
  // Now check if the target file exists already
  id=rename(filename,ad_desname);
  if (id) {
    if (errno==35) {
      eprint("File exists already: (%s)\r\n",ad_desname);
      errors++;
    }
    else {
      eprint("Cannot create target file (%s) [reason: %d]\r\n",ad_desname,errno);
      errors++;
    }
  }
  else {
    if (y_part==0) 
      print("Decoded:  %s --> %s\r\n",fname,ad_desname);
    if (y_part)    
      print("Decoded multipart --> %s\r\n",ad_desname);
  }
 abortus:
  goto loop;
  
  
 eof1:
  fclose(fIn);
  return(0);
}

/*------------------------------------------------------------------------------*/

void print_usage(char *exec) {
  print("\r\n");
  printf("Usage: %s [file(s)] options \n\n", exec);
  printf("Options:\n");
  printf("-a\tShow debug output\n\n");

}

/*------------------------------------------------------------------------------*/

int main(int argc, char * argv[]) {
  int id, file_proc;
  char mask[260];
  char * cp;
  
  errors=0;  //  Counter for batch decoder errors
  errfiles=0;
  errparts=0;
  
  aDsp=0;            // all debug display
  datadir[0]=0;      // Target directory is actual directory
  ad_despath[260]=0;
  
  ad_cleanmode=0;    // No special modes for standalone decoder
  ad_crcmode=0;
  
  print("\r\n");
  print("yDec-v1 (14.Nov.2001) - Public Domain - by Juergen Helbing - www.yenc.org\n");
  print("Linux/Unix porting done by Michael Gaskins (1/12/2002)\n\n");
  
  if (argc < 2)  {
    print("Missing parameter (filename or filemask required) !\r\n");
    print_usage(argv[0]);
    return 1;
  }

  // Now find all the mesages and decode them
  else {
    file_proc = argc - 1;
    if ((!strcmp(argv[file_proc],"/a")) || (!strcmp(argv[file_proc], "-a"))) { 
      aDsp=1;
      file_proc--;
    }
    /* 
       this will process multiple files entered sequentially at the prompt.  
       this is also how the wildcard feature is processed in Unices (by the 
       shell expanding the arguments) so it was somewhat of a required change 
       from the original Win32 code
    */
    while (file_proc > 0) {
      printf ("Now processing: %s\n", argv[file_proc]);
      DecodeFile(argv[file_proc--]);
    }
  }
  

  /* All info on errors that might have occured is now presented */
  
  if (errors|errfiles|errparts) {
    print("\r\n");
    if (errors) 
      print("%d error(s) during decoding.\r\n",errors);
    if (errfiles) 
      print("%d file(s) with errors stored to disk\r\n",errfiles);
    if (errparts) 
      print("%d corrupt part(s) detected.\r\n",errparts);
    
    return(0);
  }
}

/*------------------------------------------------------------------------------*/
